'''
Copied from python 3.8 functools for 3.7 support
'''

from functools import singledispatch

WRAPPER_ASSIGNMENTS = ('__module__', '__name__', '__qualname__', '__doc__',
                       '__annotations__')
WRAPPER_UPDATES = ('__dict__',)


def update_wrapper(wrapper,
                   wrapped,
                   assigned = WRAPPER_ASSIGNMENTS,
                   updated = WRAPPER_UPDATES):
    """Update a wrapper function to look like the wrapped function
       wrapper is the function to be updated
       wrapped is the original function
       assigned is a tuple naming the attributes assigned directly
       from the wrapped function to the wrapper function (defaults to
       functools.WRAPPER_ASSIGNMENTS)
       updated is a tuple naming the attributes of the wrapper that
       are updated with the corresponding attribute from the wrapped
       function (defaults to functools.WRAPPER_UPDATES)
    """
    for attr in assigned:
        try:
            value = getattr(wrapped, attr)
        except AttributeError:
            pass
        else:
            setattr(wrapper, attr, value)
    for attr in updated:
        getattr(wrapper, attr).update(getattr(wrapped, attr, {}))
    # Issue #17482: set __wrapped__ last so we don't inadvertently copy it
    # from the wrapped function when updating __dict__
    wrapper.__wrapped__ = wrapped
    # Return the wrapper so this can be used as a decorator via partial()
    return wrapper

# Descriptor version
class singledispatchmethod:
    """Single-dispatch generic method descriptor.
    Supports wrapping existing descriptors and handles non-descriptor
    callables as instance methods.
    """

    def __init__(self, func):
        if not callable(func) and not hasattr(func, "__get__"):
            raise TypeError(f"{func!r} is not callable or a descriptor")

        self.dispatcher = singledispatch(func)
        self.func = func

    def register(self, cls, method=None):
        """generic_method.register(cls, func) -> func
        Registers a new implementation for the given *cls* on a *generic_method*.
        """
        return self.dispatcher.register(cls, func=method)

    def __get__(self, obj, cls=None):
        def _method(*args, **kwargs):
            class_ = args[0] if isinstance(args[0], type) else args[0].__class__
            method = self.dispatcher.dispatch(class_)
            return method.__get__(obj, cls)(*args, **kwargs)

        _method.__isabstractmethod__ = self.__isabstractmethod__
        _method.register = self.register
        update_wrapper(_method, self.func)
        return _method

    @property
    def __isabstractmethod__(self):
        return getattr(self.func, '__isabstractmethod__', False)




import pytz
import numbers
import pandas as pd
from pydantic import ValidationError
from pydantic.fields import ModelField, FieldInfo
from typing import Generic, Any

from datetime import datetime

from rframe.utils import singledispatchmethod
from .types import LabelType


class BaseIndex(FieldInfo):
    __slots__ = FieldInfo.__slots__ + ('name', 'schema', 'label_type', 'nullable')

    def __init__(self, default: Any = ..., **kwargs: Any) -> None:
        self.nullable = kwargs.pop('nullable', True)
        super().__init__(default, **kwargs)
        self.name = 'index'
        self.schema = None
        self.label_type = Any
        
    def __set_name__(self, owner, name):
        self.name = name
        self.schema = owner
        self.label_type = owner.__fields__[name].type_

    @property
    def names(self):
        return [self.name]

    def validate_label(self, label):
        if isinstance(label, slice):
            start = self.validate_label(label.start)
            stop = self.validate_label(label.stop)
            step = self.validate_label(label.step)
            if start is None and stop is None:
                label = None
            else:
                return slice(start, stop, step)

        if label is None and self.nullable:
            return label

        if isinstance(label, dict) and self.label_type is not dict:
            return {k: self.validate_label(val) for k,val in label.items()}

        if isinstance(label, list) and self.label_type is not list:
            return [self.validate_label(val) for val in label]

        if isinstance(label, tuple) and self.label_type is not tuple:
            return tuple(self.validate_label(val) for val in label)

        label = self.coerce(label)
        
        if not isinstance(label, self.label_type):
            raise TypeError(f'{self.name} must be of type {self.label_type}')

        return label

    def coerce(self, label):
        if isinstance(label, self.label_type):
            return label
        
        label =  self._coerce(self.label_type, label)

        return label

    @singledispatchmethod
    def _coerce(self, type_, value):
        return type_(value)

    @_coerce.register(datetime)
    def _coerce_datetime(self, type_, value):
        if isinstance(value, datetime):
            if value.tzinfo is not None and value.tzinfo.utcoffset(value) is not None:
                value = value.astimezone(pytz.utc)
            else:
                value = value.replace(tzinfo=pytz.utc)
            return value
        unit = self.unit if isinstance(value, numbers.Number) else None
        utc = getattr(self, 'utc', True)
        value = pd.to_datetime(value, utc=utc, unit=unit).to_pydatetime()
        return self._coerce_datetime(type_, value)

    def to_pandas(self, label):
        return label

    def from_pandas(self, label):
        return label

    def reduce(self, docs, labels):
        return docs

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name},type={self.label_type})"






import pandas as pd

from typing import Iterable, Mapping
from nbformat import ValidationError

from .types import Interval
from .base import BaseIndex


class IntervalIndex(BaseIndex):
    __slots__ = BaseIndex.__slots__ + ('left_name', 'right_name', 'closed')

    def __set_name__(self, owner, name):
        self.name = name
        self.schema = owner
        if owner.__fields__[name].type_ is not Interval:
            raise ValidationError(f'{name} is a interval index but is not of type Interval')
        sub_fields = owner.__fields__[name].sub_fields
        if not sub_fields:
            raise ValidationError(f'You must specify a label type for the {name} index')
        self.label_type = sub_fields[0].type_

    def __init__(self, closed='right', **kwargs):
        super().__init__(**kwargs)
        
        self.closed = closed

    def to_pandas(self, label):
        if isinstance(label, Mapping):
            left = pd.to_datetime(label['left'])
            right = pd.to_datetime(label['right'])
            
        elif isinstance(label, Interval):
            left = pd.to_datetime(label.left)
            right = pd.to_datetime(label.right)

        elif isinstance(label, Iterable) and len(label)==2:
            left = pd.to_datetime(label[0])
            right = pd.to_datetime(label[1])
        else:
            raise TypeError(f"{self.name} must be a Mapping,Interval"
             "or Iterable of length 2, got {type(label)} instead.")

        label = pd.Interval(left, right, closed=self.closed)
        return label

    def from_pandas(self, label):
        if isinstance(label, pd.Interval):
            label = Interval(left=label.left, right=label.right)
        return label


from abc import ABC, abstractmethod
from ..utils import singledispatchmethod


class BaseDataQuery(ABC):

    @abstractmethod
    def apply(self):
        pass

    @abstractmethod
    def __add__(self, other):
        pass


class DatasourceIndexer(ABC):
    INTERFACES = {}

    def __init__(self, source) -> None:
        self.source = source

    @classmethod
    def register_indexer(cls, source_type, indexer = None):
        def wrapper(indexer):
            if isinstance(source_type, tuple):
                for t in source_type:
                    cls.register(t, indexer)
                return indexer

            if source_type in cls.INTERFACES:
                raise ValueError(f'Interface for source {source_type} already registered.')
            cls.INTERFACES[source_type] = indexer
            return indexer
        return wrapper(indexer) if indexer is not None else wrapper

    @classmethod
    def from_source(cls, source, *args, **kwargs):
        type_ = type(source)
        if type_ in cls.INTERFACES:
            return cls.INTERFACES[type_](source, *args, **kwargs)
        raise NotImplementedError(f'No implementation for data source of type {type(source)}')

    @abstractmethod
    def compile_query(self, index, label):
        pass
    
    def insert(self, source, doc):
        raise NotImplementedError



import datetime
import pandas as pd
from typing import Generic, Mapping, TypeVar
from numpy import isin

from pydantic import BaseModel, ValidationError

LabelType = TypeVar('LabelType', int, str, datetime.datetime)


class Interval(BaseModel, Generic[LabelType]):
    left: LabelType
    right: LabelType = None

    @classmethod
    def __get_validators__(cls):
        yield cls.validate_field

    @classmethod
    def validate_field(cls, v, field):
        
        if isinstance(v, tuple):
            left, right = v
        elif isinstance(v, Mapping):
            left = v.get('left', None)
            right = v.get('right', None)
        elif isinstance(v, (pd.Interval, Interval)):
            left = v.left
            right = v.right
        else:
            left, right = v, v
            
        if field.sub_fields:
            left, error = field.sub_fields[0].validate(left, {}, loc='LabelType')
            if error:
                raise ValidationError([error])
            if right is not None:
                right, error = field.sub_fields[0].validate(right, {}, loc='LabelType')
                if error:
                    raise ValidationError([error])
        return cls(left=left, right=right)






import re
import toolz
from typing import Dict, ClassVar, List
from pydantic import BaseModel
from pydantic.fields import ModelField, FieldInfo

from .indexes import BaseIndex, Index, MultiIndex
from .indexers import get_indexer

def camel_to_snake(name):
  name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
  return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()


class InsertionError(Exception):
    pass


class BaseSchema(BaseModel):
    _name: str = ''
    
    def __init_subclass__(cls) -> None:
        if 'name' not in cls.__dict__:
            cls.name = camel_to_snake(cls.__name__)
        index_fields = cls.get_index_fields()
        indexes = []
        for name, field in index_fields.items():
            index = field.field_info
            index.__set_name__(cls, name)
            indexes.append(index)

        if len(indexes) == 1:
            index = indexes[0]
        else:
            index = MultiIndex(*indexes)

        cls.index = index

    @classmethod
    def field_info(cls) -> Dict[str,FieldInfo]:
        return {name: field.field_info 
                for name, field in cls.__fields__.items()}

    @classmethod
    def get_index_fields(cls) -> Dict[str,ModelField]:
        fields = {}
        for name, field in cls.__fields__.items():
            if isinstance(field.field_info, BaseIndex):
                fields[name] = field
        return fields

    @classmethod
    def get_column_fields(cls) -> Dict[str,ModelField]:
        fields = {}
        for name, field in cls.__fields__.items():
            if not isinstance(field.field_info, BaseIndex):
                fields[name] = field
        return fields

    @classmethod
    def get_index(cls) -> BaseIndex:
        index_fields = cls.get_index_fields()
        indexes = []
        for name, field in index_fields.items():
            index = field.field_info
            index.__set_name__(cls, name)
            indexes.append(index)

        if len(indexes) == 1:
            index = indexes[0]
        else:
            index = MultiIndex(*indexes)

        return index

    @classmethod
    def index_for(cls, name):
        field_info = cls.field_info().get(name, None)
        if not isinstance(field_info, BaseIndex):
            field_info = Index()
        field_info.__set_name__(cls, name)
        return field_info

    @property
    def index_labels(self):
        data = self.dict()
        return {k: data[k] for k in self.get_index_fields()}

    @classmethod
    def compile_query(cls, datastore, **labels)-> List['BaseSchema']:
        indexes = [cls.index_for(name) for name in labels]
        if len(indexes) == 1:
            index = indexes[0]
            label = labels[index.name]
        else:
            index = MultiIndex(*indexes)
            label = labels

        label = index.validate_label(label)
        indexer = get_indexer(datastore)
        
        return indexer.compile_query(index, label)

    @classmethod
    def find_raw(cls, datastore, **labels)-> List['BaseSchema']:
        indexes = [cls.index_for(name) for name in labels]
        if len(indexes) == 1:
            index = indexes[0]
            label = labels[index.name]
        else:
            index = MultiIndex(*indexes)
            label = labels

        label = index.validate_label(label)
        indexer = get_indexer(datastore)
        
        query =  indexer.compile_query(index, label)
        docs = query.apply(datastore)
        docs = index.reduce(docs, labels)
        return docs

    @classmethod
    def find(cls, datastore, **labels)-> List['BaseSchema']:
        docs = cls.find_raw(datastore, **labels)
        if not docs:
            return []
        
        docs = [cls(**doc) for doc in docs]

        return docs

    @classmethod
    def ensure_index(cls, datastore):
        pass

    def save(self, datastore):
        indexer = get_indexer(datastore)
        existing = self.find(datastore, **self.index_labels)
        if existing:
            existing[0].pre_update(datastore, self)
        else:
            self.pre_insert(datastore)
        return indexer.insert(datastore, self)

    def pre_insert(self, datastore):
        pass

    def pre_update(self, datastore, new):
        pass
    
    def same_values(self, other):
        return all([getattr(self, attr) == getattr(other, attr)
                     for attr in self.get_column_fields()])

    def pandas_dict(self):
        data = self.dict()
        for name, label in self.index_labels.items():
            index = self.index_for(name)
            data[name] = index.to_pandas(label)
        return data



from typing import Callable, Union, Generic
import datetime
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from rframe.indexes.types import LabelType

from .base import BaseIndex
from ..utils import singledispatchmethod, singledispatch


def nn_interpolate(x, xs, ys):
    idx = np.argmin(np.abs(x-np.array(xs)))
    return ys[idx]

@singledispatch
def interpolater(x, xs, ys, kind='linear'):
    raise f"Interpolation on type {type(x)} is not supported."

@interpolater.register(float)
@interpolater.register(int)
def interpolate_number(x, xs, ys, kind='linear'):
    if isinstance(ys[0], (float, int)):
        func = interp1d(xs, ys, fill_value=(ys[0], ys[-1]),
                        bounds_error=False, kind=kind)
        return func(x).item()
    return nn_interpolate(x,xs,ys)

@interpolater.register(datetime.datetime)
def interpolate_datetime(x, xs, ys, kind='linear'):
    xs = [x.timestamp() for x in xs]
    x = x.timestamp()
    if isinstance(ys[0], (float, int)):
        return interpolate_number(x, xs, ys, kind=kind)
    return nn_interpolate(x,xs,ys)


class InterpolatingIndex(BaseIndex):
    kind: str = 'linear'
    neighbours: int = 1
    inclusive: bool = False
    extrapolate: Union[bool,Callable] = False

    def __init__(self, kind='linear', neighbours=1, 
                inclusive=False, extrapolate=False, **kwargs):
        super().__init__(**kwargs)
        self.kind = kind
        self.neighbours = neighbours
        self.inclusive = inclusive
        self.extrapolate = extrapolate

    def can_extrapolate(self, index):
        if callable(self.extrapolate):
            return self.extrapolate(index)
        return self.extrapolate

    def reduce(self, docs, label):
        if label is None:
            return docs

        if isinstance(label, list):
            return [d for val in label 
                    for d in self.reduce(docs, val)]

        label = self.validate_label(label)
        x = label.timestamp() if isinstance(label, datetime.datetime) else label

        xs = [self.validate_label(d[self.name]) for d in docs]

        # just convert all datetimes to timestamps to avoid complexity
        # FIXME: maybe properly handle timezones instead
        xs = [xi.timestamp() if isinstance(xi, datetime.datetime) else xi
                for xi in xs]

        new_document = dict(nn_interpolate(x, xs, docs))
        new_document[self.name] = label
        
        if len(xs)>1 and max(xs)>=x>=min(xs):
            for yname in self.schema.get_column_fields():
                ys = [d[yname] for d in docs]
                new_document[yname] = interpolater(x, xs, ys, kind=self.kind)
            return [new_document]

        if (x>max(xs) and self.can_extrapolate(new_document)) or x==max(xs):
            return [new_document]
        
        return []



from pydantic import ValidationError
from pydantic.fields import ModelField
import datetime
from typing import TypeVar, Generic

from rframe.utils import singledispatchmethod
from .base import BaseIndex
from .types import LabelType
# IndexType = TypeVar('IndexType', int, str, datetime.datetime)


class Index(BaseIndex):
    pass
    



from .base import DatasourceIndexer
from .mongo import MongoIndexer
from ..utils import singledispatch


def get_indexer(source) -> DatasourceIndexer:
    return DatasourceIndexer.from_source(source)



"""Main module."""

from datetime import datetime
import re
from pydantic.typing import NoneType
import pymongo
import pandas as pd

from typing import Any, Dict, List, Type, Union, Tuple

from rframe.indexers import get_indexer
from .schema import BaseSchema, InsertionError
from .utils import singledispatchmethod
from .indexers import get_indexer

IndexLabel = Union[int,float,datetime,str,slice,NoneType,List]


class RemoteFrame:
    '''Implement basic indexing features similar to a pandas dataframe
    but operates on an arbitrary storage backend
    '''
    schema: Type[BaseSchema]
    db: Any

    def __init__(self, schema: Type[BaseSchema], db: Any, **kwargs) -> None:
        if isinstance(db, str):
            if db.startswith('mongodb'):
                dbname = kwargs.pop('dbname', 'remote_dataframes')
                db = pymongo.MongoClient(db, **kwargs)[dbname]
            elif db.endswith('.csv'):
                db = pd.read_csv(db)
            elif db.endswith('.pkl'):
                db = pd.read_pickle(db)
            elif db.endswith('.pq'):
                db = pd.read_parquet(db)
            else:
                raise TypeError("Unsupported database type")
        self.schema = schema
        self.db = db

    @classmethod
    def from_mongodb(cls, schema, url, dbname='remote_dataframes', **kwargs):
        import pymongo
        db = pymongo.MongoClient(url, **kwargs)[dbname]
        return cls(schema, db)

    @property
    def name(self):
        return self.schema.name
    
    @property
    def indexer(self):
        return get_indexer(self.db)

    @property
    def columns(self):
        return list(self.schema.get_column_fields())

    @property
    def index(self):
        return self.schema.get_index()

    @property
    def loc(self):
        return LocIndexer(self)

    @property
    def at(self):
        return AtIndexer(self)

    def sel_records(self, *args: IndexLabel, **kwargs: IndexLabel) -> List[dict]:
        '''Queries the DB and returns the results as a list of dicts
        '''
        index_fields = self.index.names
        labels = {name: lbl for name, lbl in zip(index_fields, args)}
        labels.update(kwargs)
        docs = self.schema.find(self.db, **labels)
        docs = [doc.pandas_dict() for doc in docs]
        return docs
        # return self.index.query_db(self.db, *args, **kwargs)

    def sel_record(self, *args: IndexLabel, **kwargs: IndexLabel) -> dict:
        '''Return a single dict
        '''
        records = self.sel_records(*args, **kwargs)
        if records:
            return records[0]       
        raise KeyError('Selection returned no records.')

    def head(self, n=10) -> pd.DataFrame:
        '''Return first n documents as a pandas dataframe
        '''
        docs = self.schema.index.head(self.db, n)
        index_fields = self.schema.index_names()
        df = pd.DataFrame(docs, columns=self.schema.all_fields())
        idx = [c for c in index_fields if c in df.columns]
        return df.sort_values(idx).set_index(idx)

    def sel(self, *args: IndexLabel, **kwargs: IndexLabel) -> pd.DataFrame:
        '''select a subset of the data
        returns a pandas dataframe
        '''
        docs = self.sel_records(*args, **kwargs)
        df = pd.DataFrame(docs, columns=self.index.names+self.columns)

        idx = [c for c in self.schema.get_index_fields() if c in df.columns]
        return df.sort_values(idx).set_index(idx)

    def set(self, *args: IndexLabel, **kwargs: IndexLabel) -> BaseSchema:
        '''Insert data by index
        '''
        labels = {name: lbl for name, lbl in zip(self.index.names, args)}
        kwargs.update(labels)
        doc = self.schema(**kwargs)
        return doc.save(self.db, doc)

    def insert(self, records: Union[pd.DataFrame,List[dict]]) -> Tuple[List[dict],List[dict],List[dict]]:
        ''' Insert multiple records into the DB
        '''
        if isinstance(records, pd.DataFrame):
            records = records.reset_index().to_dict(orient='records')
        succeeded = []
        failed = []
        errors = []
        for record in records:
            doc = self.schema(**record)
            try:
                doc.save(self.db)
                succeeded.append(doc.dict())
            except InsertionError as e:
                failed.append(doc.dict())
                errors.append(str(e))

        return succeeded, failed, errors

    def __getitem__(self, index: Tuple[IndexLabel,...]) -> 'RemoteSeries':
        if isinstance(index, str) and index in self.columns:
            return RemoteSeries(self, index)
        if isinstance(index, tuple) and index[0] in self.columns:
            return RemoteSeries(self, index[0])[index[1:]]
        raise KeyError(f'{index} is not a dataframe column.')

    def __call__(self, column: str, **index: IndexLabel) -> pd.DataFrame:
        index = tuple(index.get(k, None) for k in self.index.names)
        return self.at[index, column]

    def __dir__(self) -> List[str]:
        return self.columns + super().__dir__()

    def __getattr__(self, name: str) -> 'RemoteSeries':
        if name != 'columns' and name in self.columns:
            return self[name]
        raise AttributeError(name)

    def __repr__(self) -> str:
        return (f"RemoteDataFrame("
               f"index={self.index.names},"
               f"columns={self.columns})")

class RemoteSeries:
    obj: RemoteFrame
    column: str

    def __init__(self, obj: RemoteFrame, column: str) -> None:
        self.obj = obj
        self.column = column

    def __getitem__(self, index: Union[IndexLabel,Tuple[IndexLabel,...]]) -> pd.DataFrame:
        if not isinstance(index, tuple):
            index = (index,)
        return self.obj.sel(*index)[self.column]

    def sel(self, *args: IndexLabel, **kwargs: IndexLabel) -> pd.DataFrame:
        df = self.obj.sel(*args, **kwargs)
        return df[self.column]

    def sel_values(self, *args: IndexLabel, **kwargs: IndexLabel) -> List[Any]:
        docs = self.obj.sel_records(*args, **kwargs)
        return [doc[self.column] for doc in docs]

    def sel_value(self, *args: IndexLabel, **kwargs: IndexLabel) -> Any:
        values = self.sel_values(*args, **kwargs)
        if values:
            return values[0]
        raise KeyError('Selection returned no values.')

    def set(self,  *args: IndexLabel, **kwargs: IndexLabel):
        raise InsertionError('Cannot set values on a RemoteSeries object,'
                             'use the RemoteDataFrame.')

    def __repr__(self) -> str:
        return (f"RemoteSeries(index={self.obj.index.names},"
                f"column={self.column})")


class Indexer:
    def __init__(self, obj: RemoteFrame):
        self.obj = obj


class LocIndexer(Indexer):

    def __call__(self, *args: IndexLabel, **kwargs: IndexLabel) -> pd.DataFrame:
        return self.obj.sel(*args, **kwargs)

    def __getitem__(self, index: Tuple[IndexLabel]) -> pd.DataFrame:
        columns = None
        
        if isinstance(index, tuple) and len(index) == 2:
            index, columns = index
            if not isinstance(columns, list):
                columns = [columns]
            if not all([c in self.obj.columns for c in columns]):
                if not isinstance(index, tuple):
                    index = (index,)
                index = index + tuple(columns)
                columns = None

        elif isinstance(index, tuple) and len(index) == len(self.obj.columns)+1:
            index, columns = index[:-1], index[-1]

        if not isinstance(index, tuple):
            index = (index,)
        
        df = self.obj.sel(*index)
    
        if columns is not None:
            df = df[columns]

        return df

    def __setitem__(self, key: str, value: Union[dict,BaseSchema]) -> BaseSchema:
        if not isinstance(key, tuple):
            key = (key,)
        
        if isinstance(value, self.obj.schema):
            value  = value.dict()

        if not isinstance(value, dict):
            value = {'value': value}

        return self.obj.set(*key, **value)


class AtIndexer(Indexer):

    def __getitem__(self, key: Tuple[Tuple[IndexLabel, ...],str]) -> Any:
        
        if not (isinstance(key, tuple) and len(key)==2):
            raise KeyError('ill-defined location. Specify '
                           '.at[index,column] where index can be a tuple.')
        
        index, column = key

        if column not in self.obj.columns:
            raise KeyError(f'{column} not found. Valid columns are: {self.obj.columns}')
        
        if not isinstance(index, tuple):
            index = (index,)

        if any([isinstance(idx, (slice, list, type(None))) for idx in index]):
            raise KeyError(f'{index} is not unique index.')

        if len(index)<len(self.obj.index.names):
            KeyError(f'{index} is an under defined index.')

        return self.obj[column].sel_value(*index)




import pymongo
from pymongo.collection import Collection

from .base import BaseDataQuery, DatasourceIndexer
from ..utils import singledispatchmethod
from ..indexes import Index, InterpolatingIndex, IntervalIndex, MultiIndex


class MongoAggregation(BaseDataQuery):
    pipeline: list

    def __init__(self, pipeline):
        self.pipeline = pipeline
  
    def apply(self, collection: Collection):
        if not isinstance(collection, Collection):
            raise TypeError(f'collection must be a pymongo Collection, got {type(collection)}.')

        return list(collection.aggregate(self.pipeline))

    def logical_and(self, other):
        return MongoAggregation(self.pipeline+other.pipeline)

    def __add__(self, other):
        return self.logical_and(other)

    def __and__(self, other):
        return self.logical_and(other)


@DatasourceIndexer.register_indexer(pymongo.collection.Collection)
class MongoIndexer(DatasourceIndexer):

    @singledispatchmethod
    def compile_query(self, index, label):
        raise NotImplementedError(f'{self.__class__.__name__} does not support {type(index)} indexes.')
    
    @compile_query.register(list)
    @compile_query.register(tuple)
    @compile_query.register(MultiIndex)
    def multi_query(self, index, labels):
        if isinstance(index, MultiIndex):
            index = index.indexes
            labels = labels.values()
        agg = MongoAggregation([])
        for idx,label in zip(index, labels):
            agg = agg + self.compile_query(idx, label)
        return agg

    @compile_query.register(Index)
    @compile_query.register(str)
    def get_simple_query(self, index, label):
        name = index.name if isinstance(index, Index) else index

        if isinstance(label, slice):
            # support basic slicing, this will only work
            # for values that are comparable with the
            #  $gt/$lt operators
            start = label.start
            stop = label.stop
            step = label.step
            if step is None:
                label = {}
                if start is not None:
                    label['$gte'] = start
                if stop is not None:
                    label['$lt'] = stop
                if not label:
                    label = None
            else:
                label = list(range(start, stop, step))

        if isinstance(label, list):
            # support querying multiple values
            # in the same request
            label = {'$in': label}

        pipeline = []
        if label is not None:
            pipeline.append({'$match': {name: label} })

        return MongoAggregation(pipeline)

    @compile_query.register(InterpolatingIndex)
    def build_interpolation_query(self, index, label):
        '''For interpolation we match the values directly before and after
        the value of interest. For each value we take the closest document on either side.
        '''
        if label is None:
            return MongoAggregation([])

        if not isinstance(label, list):
            label = [label]

        queries = {f'agg{i}': mongo_closest_query(index.name, value) for i,value in enumerate(label)}
        pipeline = [
                {
                    # support multiple independent aggregations
                    # using the facet feature 
                    '$facet': queries,
                },

                # Combine results of all aggregations 
                {
                    '$project': {
                        'union': {
                            '$setUnion': [f'${name}' for name in queries],
                            }
                            }
                },
                # we just want a single list of documents
                {
                    '$unwind': '$union'
                },
                # move list of documents to the root of the result
                # so we just get a nice list of documents
                {
                    '$replaceRoot': { 'newRoot': "$union" }
                },
            ]
        return MongoAggregation(pipeline)     

    @compile_query.register(IntervalIndex)
    def build_interval_query(self, index, intervals):
        '''Query overlaping documents with given interval, supports multiple 
        intervals as well as zero length intervals (left==right)
        multiple overlap queries are joined with the $or operator 
        '''
        if not isinstance(intervals, list):
            intervals = [intervals]

        queries = []
        for interval in intervals:
            if interval is None:
                continue
            if isinstance(interval, tuple) and all([i is None for i in interval]):
                continue
            query = mongo_overlap_query(index, interval)
            if query:
                queries.append(query)

        if queries:
            pipeline = [
            {
                '$match':  {
                    # support querying for multiple values
                    # in a single pipeline
                    '$or': queries,
                }
            }
            ]
        else:
            pipeline = []
        
        return MongoAggregation(pipeline)

    def insert(self, collection: Collection, doc):
        '''We want the client logic to be agnostic to
        whether the value being replaced is actually stored in the DB or
        was inferred from e.g interpolation. 
        The find_one_and_replace(upsert=True) logic is the best match
        for the behavior we want even though it wasts an insert operation
        when a document already exists.
        FIXME: Maybe we can optimize this with an pipeline to
        avoid replacing existing documents with a copy.
        '''
        from rframe.schema import InsertionError

        try:
            doc = collection.find_one_and_update(doc.index_labels,
                        {'$set': doc.dict()}, projection={'_id': False},
                        upsert=True, return_document=pymongo.ReturnDocument.AFTER)
            return doc
        except Exception as e:
            raise InsertionError(f"Mongodb has rejected this insertion:\n {e} ")

def mongo_overlap_query(index, interval):
    '''Builds a single overlap query
    Intervals with one side equal to null are treated as extending to infinity in
    that direction.
    Supports closed or open intervals as well as infinite intervals

    Overlap definition:
    The two intervals (L,R) and (l,r) overlap iff L<r and l<R
    The operator < is replaced with <= when the interval is closed on that side.
    Where if L/l are None, they are treated as -inf
    and if R/r are None, the are treated as inf 
    '''

    # Set the appropriate operators depending on if the interval
    # is closed on one side or both
    closed = getattr(index, 'closed', 'right')
    gt_op = '$gte' if closed in ['right', 'both'] else '$gt'
    lt_op = '$lte' if closed in ['left', 'both'] else '$lt'

    # handle different kinds of interval definitions
    if isinstance(interval, tuple):
        left, right = interval
    elif isinstance(interval, slice):
        left, right = interval.start, interval.stop
    else:
        left = right = interval
    
    # Some conditions may not apply if the query interval is None
    # on one or both sides
    conditions = []
    if left is not None:
        conditions.append(
                {
                    '$or': [
                        # if the right side of the queried interval is
                        # None, treat it as inf
                        {f'{index.name}.right': None},
                        {f'{index.name}.right': {gt_op: left}},
                        ]
                }
            )
    if right is not None:
        conditions.append(
                {
                    '$or': [{f'{index.name}.left': None},
                            {f'{index.name}.left': {lt_op: right}}]
                }
            )
    if conditions:
        return {
                '$and': conditions,
            }
    else:
        return {}

def mongo_closest_query(name, value):
    return [
        {
            '$addFields': {
                # Add a field splitting the documents into
                # before and after the value of interest
                '_after': {'$gte': [f'${name}', value]},

                # Add a field with the distance to the value of interest
                '_diff': {'$abs': {'$subtract': [value, f'${name}']}},        
                }
        },
        {
            # sort in ascending order by distance
            '$sort': {'_diff': 1},
        },
        {
            # first group by whether document is before or after the value
            # the take the first document in each group
            '$group' : { '_id' : '$_after', 'doc': {'$first': '$$ROOT'},  }
        },
        {
            # make the documents the new root, discarding the groupby value
            "$replaceRoot": { "newRoot": "$doc" },
        },
        {
            # drop the extra fields, they are no longer needed
            '$project': {'_diff':0, '_after':0 },
        },
    ]



from numpy import isin
import toolz
from .base import BaseIndex

import collections


class FrozenDict(collections.Mapping):
    '''https://stackoverflow.com/questions/2703599/what-would-a-frozen-dict-be
    '''
    def __init__(self, *args, **kwargs):
        self._d = dict(*args, **kwargs)
        self._hash = None

    def __iter__(self):
        return iter(self._d)

    def __len__(self):
        return len(self._d)

    def __getitem__(self, key):
        return self._d[key]

    def __hash__(self):
        # It would have been simpler and maybe more obvious to 
        # use hash(tuple(sorted(self._d.iteritems()))) from this discussion
        # so far, but this solution is O(n). I don't know what kind of 
        # n we are going to run into, but sometimes it's hard to resist the 
        # urge to optimize when it will gain improved algorithmic performance.
        if self._hash is None:
            hash_ = 0
            for pair in self.items():
                hash_ ^= hash(pair)
            self._hash = hash_
        return self._hash

def hashable_doc(doc):
    if isinstance(doc, dict):
        doc = {k: hashable_doc(v) for k,v in doc.items()}
        return FrozenDict(doc)
    return doc

def unhashable_doc(doc):
    if isinstance(doc, FrozenDict):
        doc = {k: unhashable_doc(v) for k,v in doc.items()}
        return dict(doc)
    return doc

class MultiIndex(BaseIndex):
    _indexes: list

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(**kwargs)
        indexes = list(args)
 
        for i, index in enumerate(indexes):
            if index.name  in ['', 'index']:
                index.name = f'index_{i}'
        self._indexes = indexes

    @property
    def indexes(self):
        return getattr(self, '_indexes', [])[:]

    @property
    def names(self):
        return [index.name for index in self.indexes]

    def validate_label(self, label: dict) -> dict:
        indexes = {index.name: index for index in self.indexes}
        return {k: indexes[k].validate_label(v)
                for k,v in label.items()}

    def reduce(self, documents, labels):
        if not documents:
            return documents
        documents = [hashable_doc(doc) for doc in documents]

        keys = set(index.name for index in self.indexes)
        keys = keys.intersection(documents[0])

        for index in self.indexes:
            if index.name not in labels:
                continue
            others = [k for k in keys if k not in index.names]
            if not others:
                continue
            reduced_documents = []
            for _,docs in toolz.groupby(others, documents).items():
                label = labels[index.name]
                reduced  = index.reduce(docs, label)
                reduced_documents.extend(reduced)
            documents = reduced_documents
        documents = [unhashable_doc(doc) for doc in documents]
        return documents

    def __repr__(self):
        return f"MultiIndex({self.indexes})"



"""Top-level package for rframe."""

__author__ = """Yossi Mosbacher"""
__email__ = 'joe.mosbacher@gmail.com'
__version__ = '0.1.0'


from .rframe import RemoteFrame
from .schema import BaseSchema
from .indexes import *



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rframe', 'rframe.indexers', 'rframe.indexes', 'tests']

package_data = \
{'': ['*']}

install_requires = \
['click',
 'pandas>=1.4.0,<2.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'pymongo>=4.0.1,<5.0.0',
 'pytoolz>=0.1.7,<0.2.0']

entry_points = \
{'console_scripts': ['rframe = rframe.cli:main']}

setup_kwargs = {
    'name': 'rframe',
    'version': '0.1.0',
    'description': 'Top-level package for rframe.',
    'long_description': '======\nrframe\n======\n\n\n.. image:: https://img.shields.io/pypi/v/rframe.svg\n        :target: https://pypi.python.org/pypi/rframe\n\n.. image:: https://img.shields.io/travis/jmosbacher/rframe.svg\n        :target: https://travis-ci.com/jmosbacher/rframe\n\n.. image:: https://readthedocs.org/projects/rframe/badge/?version=latest\n        :target: https://rframe.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\n\n\n\nDataframe-like indexing on database tables\n\n\n* Free software: MIT\n* Documentation: https://rframe.readthedocs.io.\n\n\nFeatures\n--------\n\n* TODO\n\nCredits\n-------\n\nThis package was created with Cookiecutter_ and the `briggySmalls/cookiecutter-pypackage`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`briggySmalls/cookiecutter-pypackage`: https://github.com/briggySmalls/cookiecutter-pypackage\n',
    'author': 'Yossi Mosbacher',
    'author_email': 'joe.mosbacher@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jmosbacher/rframe',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



from .base import BaseIndex
from .index import Index
from .interpolating_index import InterpolatingIndex
from .interval_index import IntervalIndex
from .multi_index import MultiIndex
from .types import Interval